@@ -32,13 +32,13 @@
 namespace Shiboken
 {
 
-typedef google::dense_hash_map<const void*, PyObject*> WrapperMap;
+typedef google::dense_hash_map<const void*, SbkObject*> WrapperMap;
 
 class Graph
 {
 public:
-    typedef std::list<SbkBaseWrapperType*> NodeList;
-    typedef google::dense_hash_map<SbkBaseWrapperType*, NodeList> Edges;
+    typedef std::list<SbkObjectType*> NodeList;
+    typedef google::dense_hash_map<SbkObjectType*, NodeList> Edges;
 
     Edges m_edges;
 
@@ -47,7 +47,7 @@ public:
         m_edges.set_empty_key(0);
     }
 
-    void addEdge(SbkBaseWrapperType* from, SbkBaseWrapperType* to)
+    void addEdge(SbkObjectType* from, SbkObjectType* to)
     {
         m_edges[from].push_back(to);
     }
@@ -61,7 +61,7 @@ public:
 
         Edges::const_iterator i = m_edges.begin();
         for (; i != m_edges.end(); ++i) {
-            SbkBaseWrapperType* node1 = i->first;
+            SbkObjectType* node1 = i->first;
             const NodeList& nodeList = i->second;
             NodeList::const_iterator j = nodeList.begin();
             for (; j != nodeList.end(); ++j)
@@ -71,19 +71,19 @@ public:
     }
 #endif
 
-    SbkBaseWrapperType* identifyType(void* cptr, SbkBaseWrapperType* type, SbkBaseWrapperType* baseType) const
+    SbkObjectType* identifyType(void* cptr, SbkObjectType* type, SbkObjectType* baseType) const
     {
         Edges::const_iterator edgesIt = m_edges.find(type);
         if (edgesIt != m_edges.end()) {
             const NodeList& adjNodes = m_edges.find(type)->second;
             NodeList::const_iterator i = adjNodes.begin();
             for (; i != adjNodes.end(); ++i) {
-                SbkBaseWrapperType* newType = identifyType(cptr, *i, baseType);
+                SbkObjectType* newType = identifyType(cptr, *i, baseType);
                 if (newType)
                     return newType;
             }
         }
-        return type->type_discovery ? type->type_discovery(cptr, baseType) : 0;
+        return ((type->d && type->d->type_discovery) ? type->d->type_discovery(cptr, baseType) : 0);
     }
 };
 
@@ -91,12 +91,16 @@ public:
 #ifndef NDEBUG
 static void showWrapperMap(const WrapperMap& wrapperMap)
 {
-    printf("-------------------------------\n");
-    printf("WrapperMap: %p (size: %d)\n", &wrapperMap, (int) wrapperMap.size());
+    fprintf(stderr, "-------------------------------\n");
+    fprintf(stderr, "WrapperMap: %p (size: %d)\n", &wrapperMap, (int) wrapperMap.size());
     WrapperMap::const_iterator iter;
-    for (iter = wrapperMap.begin(); iter != wrapperMap.end(); ++iter)
-        printf("key: %p, value: %p (%s)\n", iter->first, iter->second, iter->second->ob_type->tp_name);
-    printf("-------------------------------\n");
+    for (iter = wrapperMap.begin(); iter != wrapperMap.end(); ++iter) {
+        fprintf(stderr, "key: %p, value: %p (%s, refcnt: %d)\n", iter->first,
+                                                        iter->second,
+                                                        iter->second->ob_type->tp_name,
+                                                        (int) iter->second->ob_refcnt);
+    }
+    fprintf(stderr, "-------------------------------\n");
 }
 #endif
 
@@ -107,7 +111,7 @@ struct BindingManager::BindingManagerPrivate {
 
     BindingManagerPrivate() : destroying(false) {}
     void releaseWrapper(void* cptr);
-    void assignWrapper(PyObject* wrapper, const void* cptr);
+    void assignWrapper(SbkObject* wrapper, const void* cptr);
 
 };
 
@@ -118,14 +122,12 @@ void BindingManager::BindingManagerPrivate::releaseWrapper(void* cptr)
         wrapperMapper.erase(iter);
 }
 
-void BindingManager::BindingManagerPrivate::assignWrapper(PyObject* wrapper, const void* cptr)
+void BindingManager::BindingManagerPrivate::assignWrapper(SbkObject* wrapper, const void* cptr)
 {
     assert(cptr);
     WrapperMap::iterator iter = wrapperMapper.find(cptr);
     if (iter == wrapperMapper.end())
         wrapperMapper.insert(std::make_pair(cptr, wrapper));
-    else
-        iter->second = wrapper;
 }
 
 BindingManager::BindingManager()
@@ -140,15 +142,13 @@ BindingManager::~BindingManager()
 #ifndef NDEBUG
     showWrapperMap(m_d->wrapperMapper);
 #endif
-    assert(m_d->wrapperMapper.size() == 0);
-
     /* Cleanup hanging references. We just invalidate them as when
      * the BindingManager is being destroyed the interpreter is alredy
      * shutting down. */
-    WrapperMap::iterator iter = m_d->wrapperMapper.begin();
-    for (;iter != m_d->wrapperMapper.end(); ++iter)
-        invalidateWrapper(iter->second);
-
+    while (!m_d->wrapperMapper.empty()) {
+        Object::destroy(m_d->wrapperMapper.begin()->second, const_cast<void*>(m_d->wrapperMapper.begin()->first));
+    }
+    assert(m_d->wrapperMapper.size() == 0);
     delete m_d;
 }
 
@@ -159,36 +159,42 @@ BindingManager& BindingManager::instance() {
 
 bool BindingManager::hasWrapper(const void* cptr)
 {
-    return m_d->wrapperMapper.count(cptr);
+    return m_d->wrapperMapper.find(cptr) != m_d->wrapperMapper.end();
 }
-void BindingManager::registerWrapper(SbkBaseWrapper* pyobj, void* cptr)
-{
-    SbkBaseWrapperType* instanceType = reinterpret_cast<SbkBaseWrapperType*>(pyobj->ob_type);
 
-    if (instanceType->mi_init && !instanceType->mi_offsets)
-        instanceType->mi_offsets = instanceType->mi_init(cptr);
-    m_d->assignWrapper(reinterpret_cast<PyObject*>(pyobj), cptr);
-    if (instanceType->mi_offsets) {
-        int* offset = instanceType->mi_offsets;
+void BindingManager::registerWrapper(SbkObject* pyObj, void* cptr)
+{
+    SbkObjectType* instanceType = reinterpret_cast<SbkObjectType*>(pyObj->ob_type);
+    SbkObjectTypePrivate* d = instanceType->d;
+
+    if (!d)
+        return;
+
+    if (d->mi_init && !d->mi_offsets)
+        d->mi_offsets = d->mi_init(cptr);
+    m_d->assignWrapper(pyObj, cptr);
+    if (d->mi_offsets) {
+        int* offset = d->mi_offsets;
         while (*offset != -1) {
             if (*offset > 0)
-                m_d->assignWrapper(reinterpret_cast<PyObject*>(pyobj), reinterpret_cast<void*>((std::size_t) cptr + (*offset)));
+                m_d->assignWrapper(pyObj, reinterpret_cast<void*>((std::size_t) cptr + (*offset)));
             offset++;
         }
     }
 }
 
-void BindingManager::releaseWrapper(PyObject* wrapper)
+void BindingManager::releaseWrapper(SbkObject* sbkObj)
 {
-    SbkBaseWrapperType* sbkType = reinterpret_cast<SbkBaseWrapperType*>(wrapper->ob_type);
-    int numBases = sbkType->is_multicpp ? getNumberOfCppBaseClasses(wrapper->ob_type) : 1;
+    SbkObjectType* sbkType = reinterpret_cast<SbkObjectType*>(sbkObj->ob_type);
+    SbkObjectTypePrivate* d = sbkType->d;
+    int numBases = ((d && d->is_multicpp) ? getNumberOfCppBaseClasses(sbkObj->ob_type) : 1);
 
-    void** cptrs = reinterpret_cast<SbkBaseWrapper*>(wrapper)->cptr;
+    void** cptrs = reinterpret_cast<SbkObject*>(sbkObj)->d->cptr;
     for (int i = 0; i < numBases; ++i) {
         void* cptr = cptrs[i];
         m_d->releaseWrapper(cptr);
-        if (sbkType->mi_offsets) {
-            int* offset = sbkType->mi_offsets;
+        if (d && d->mi_offsets) {
+            int* offset = d->mi_offsets;
             while (*offset != -1) {
                 if (*offset > 0)
                     m_d->releaseWrapper((void*) ((std::size_t) cptr + (*offset)));
@@ -196,9 +202,10 @@ void BindingManager::releaseWrapper(PyObject* wrapper)
             }
         }
     }
+    sbkObj->d->validCppObject = false;
 }
 
-PyObject* BindingManager::retrieveWrapper(const void* cptr)
+SbkObject* BindingManager::retrieveWrapper(const void* cptr)
 {
     WrapperMap::iterator iter = m_d->wrapperMapper.find(cptr);
     if (iter == m_d->wrapperMapper.end())
@@ -208,22 +215,23 @@ PyObject* BindingManager::retrieveWrapper(const void* cptr)
 
 PyObject* BindingManager::getOverride(const void* cptr, const char* methodName)
 {
-    PyObject* wrapper = retrieveWrapper(cptr);
+    SbkObject* wrapper = retrieveWrapper(cptr);
     if (!wrapper)
         return 0;
 
-    if (SbkBaseWrapper_instanceDict(wrapper)) {
-        PyObject* method = PyDict_GetItemString(SbkBaseWrapper_instanceDict(wrapper), methodName);
+    if (wrapper->ob_dict) {
+        PyObject* method = PyDict_GetItemString(wrapper->ob_dict, methodName);
         if (method) {
-            Py_INCREF(method);
+            Py_INCREF((PyObject*)method);
             return method;
         }
     }
 
     PyObject* pyMethodName = PyString_FromString(methodName);
-    PyObject* method = PyObject_GetAttr(wrapper, pyMethodName);
+    PyObject* method = PyObject_GetAttr((PyObject*)wrapper, pyMethodName);
 
-    if (method && PyMethod_Check(method) && reinterpret_cast<PyMethodObject*>(method)->im_self == wrapper) {
+    if (method && PyMethod_Check(method)
+        && reinterpret_cast<PyMethodObject*>(method)->im_self == reinterpret_cast<PyObject*>(wrapper)) {
         PyObject* defaultMethod;
         PyObject* mro = wrapper->ob_type->tp_mro;
 
@@ -246,101 +254,36 @@ PyObject* BindingManager::getOverride(const void* cptr, const char* methodName)
     return 0;
 }
 
-
-void BindingManager::invalidateWrapper(PyObject* pyobj)
-{
-    std::list<PyObject*> objs = splitPyObject(pyobj);
-    std::list<PyObject*>::const_iterator it;
-    for(it=objs.begin(); it != objs.end(); it++)
-        invalidateWrapper(reinterpret_cast<SbkBaseWrapper*>(*it));
-}
-
-void BindingManager::invalidateWrapper(SbkBaseWrapper* wrapper)
-{
-    if (!wrapper || ((PyObject*)wrapper == Py_None) || !SbkBaseWrapper_validCppObject(wrapper))
-        return;
-
-    GilState gil; // lock the gil to assure no one is changing the value of m_d->destroying
-
-    // skip this if the object is a wrapper class and this is not a destructor call
-    if (SbkBaseWrapper_containsCppWrapper(wrapper) && !m_d->destroying) {
-        ParentInfo* pInfo = wrapper->parentInfo;
-        // this meaning the object has a extra ref and we will remove this now
-        if (pInfo && pInfo->hasWrapperRef) {
-            delete pInfo;
-            wrapper->parentInfo = 0;
-            Py_XDECREF((PyObject*) wrapper);
-        }
-        return;
-    }
-
-    SbkBaseWrapper_setValidCppObject(wrapper, false);
-    SbkBaseWrapper_setOwnership(wrapper, false);
-
-    // If it is a parent invalidate all children.
-    if (SbkBaseWrapper_hasParentInfo(wrapper)) {
-        ChildrenList::iterator it = wrapper->parentInfo->children.begin();
-        bool parentDestroying = m_d->destroying;
-        m_d->destroying = false;
-        for (; it != wrapper->parentInfo->children.end(); ++it)
-            invalidateWrapper(*it);
-        m_d->destroying = parentDestroying;
-    }
-
-    releaseWrapper(reinterpret_cast<PyObject*>(wrapper));
-}
-
-void BindingManager::invalidateWrapper(const void* cptr)
-{
-    WrapperMap::iterator iter = m_d->wrapperMapper.find(cptr);
-    if (iter != m_d->wrapperMapper.end())
-        invalidateWrapper(iter->second);
-}
-
-void BindingManager::destroyWrapper(const void* cptr)
-{
-    WrapperMap::iterator iter = m_d->wrapperMapper.find(cptr);
-    if (iter != m_d->wrapperMapper.end())
-        destroyWrapper(reinterpret_cast<SbkBaseWrapper*>(iter->second));
-}
-
-void BindingManager::destroyWrapper(SbkBaseWrapper* wrapper)
-{
-    GilState gil;
-    m_d->destroying = true;
-    invalidateWrapper(wrapper);
-    m_d->destroying = false;
-}
-
-void BindingManager::transferOwnershipToCpp(PyObject* wrapper)
-{
-    std::list<PyObject*> objs = splitPyObject(wrapper);
-    std::list<PyObject*>::const_iterator it;
-    for(it=objs.begin(); it != objs.end(); it++)
-        transferOwnershipToCpp(reinterpret_cast<SbkBaseWrapper*>(*it));
-}
-
-void BindingManager::transferOwnershipToCpp(SbkBaseWrapper* wrapper)
-{
-    if (wrapper->parentInfo)
-        Shiboken::removeParent(wrapper);
-
-    if (SbkBaseWrapper_containsCppWrapper(wrapper))
-        SbkBaseWrapper_setOwnership(wrapper, false);
-    else
-        invalidateWrapper(wrapper);
-}
-
-void BindingManager::addClassInheritance(Shiboken::SbkBaseWrapperType* parent, Shiboken::SbkBaseWrapperType* child)
+void BindingManager::addClassInheritance(SbkObjectType* parent, SbkObjectType* child)
 {
     m_d->classHierarchy.addEdge(parent, child);
 }
 
-SbkBaseWrapperType* BindingManager::resolveType(void* cptr, Shiboken::SbkBaseWrapperType* type)
+SbkObjectType* BindingManager::resolveType(void* cptr, SbkObjectType* type)
 {
-    SbkBaseWrapperType* identifiedType = m_d->classHierarchy.identifyType(cptr, type, type);
+    SbkObjectType* identifiedType = m_d->classHierarchy.identifyType(cptr, type, type);
     return identifiedType ? identifiedType : type;
 }
 
+std::set<SbkObject*> BindingManager::getAllPyObjects()
+{
+    std::set<SbkObject*> pyObjects;
+    const WrapperMap& wrappersMap = m_d->wrapperMapper;
+    WrapperMap::const_iterator it = wrappersMap.begin();
+    for (; it != wrappersMap.end(); ++it)
+        pyObjects.insert(it->second);
+
+    return pyObjects;
+}
+
+void BindingManager::visitAllPyObjects(ObjectVisitor visitor, void* data)
+{
+    WrapperMap copy = m_d->wrapperMapper;
+    for (WrapperMap::iterator it = copy.begin(); it != copy.end(); ++it) {
+        if (hasWrapper(it->first))
+            visitor(it->second, data);
+    }
+}
+
 } // namespace Shiboken
 
